// Jacobi 3D skeleton program
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "likwid.h"

#include "timing.h"

void dummy(double*, double*);

const double oos=1./6.;

void jacobi_line(double* d, const double* s,
                 const double* top, const double* bottom,
                 const double* front, const double* back, int n) {
                 int i,start=1;
#pragma vector aligned
                 for(i=1; i<n-1; ++i) {
                     d[i] = oos*(s[i-1]+s[i+1]+top[i]+bottom[i]+front[i]+back[i]);
                 }
}

int main(int argc, char** argv) {

  double wct_start,wct_end,cput_start,cput_end,runtime,r;
  int iter,sizex,sizey,sizez,i,j,k,n,t0,t1,t;
  int ofs;
  double* phi[2];

  if(argc!=4) {
    printf("Usage: %s <sizex> <sizey> <sizez>\n",argv[0]);
    exit(1);
  }


  LIKWID_MARKER_INIT;
  sizex = atoi(argv[1]);
  sizey = atoi(argv[2]);
  sizez = atoi(argv[3]);
  phi[0] = malloc((size_t)sizex*sizey*sizez*sizeof(double));
  phi[1] = malloc((size_t)sizex*sizey*sizez*sizeof(double));

  t0=0; t1=1;

  /* initialize with parallel first touch */
#pragma omp parallel private(j,k,ofs)
{
  LIKWID_MARKER_THREADINIT;
#pragma omp for schedule(static) private(j,ofs)
  for(i=0; i<sizez; ++i) {
    for(j=0; j<sizey; ++j) {
      ofs = i*sizex*sizey + j*sizex;
      for(k=0; k<sizex; ++k) {
     	phi[t1][ofs+k] = phi[t0][ofs+k] = i*j/400.0;
      }
    }
  }	  
}


  iter=1;
  runtime=0.0;
  while(runtime<0.5) {

  timing(&wct_start, &cput_start);
  for(n=0; n<iter; n++) {
#pragma omp parallel private(j,ofs)
{
    LIKWID_MARKER_START("ITER");
#pragma omp for schedule(static)
    for(i=1; i<sizez-1; ++i) {
      for(j=1; j<sizey-1; ++j) {
        ofs = i*sizex*sizey + j*sizex;
		// line update kernel
        jacobi_line(&phi[t1][ofs],&phi[t0][ofs],&phi[t0][ofs+sizex],&phi[t0][ofs-sizex],
			&phi[t0][ofs+sizex*sizey],&phi[t0][ofs-sizex*sizey],sizex);
      }
    }
    LIKWID_MARKER_STOP("ITER");
}
    dummy(phi[0],phi[1]);
    t=t0; t0=t1; t1=t;
  }	  

  timing(&wct_end, &cput_end);
  runtime = wct_end-wct_start;
  iter *= 2;
  }

  iter /= 2;

  printf("size: %d %d %d time: %lf  iter: %d  MLUP/s: %lf\n",sizex,sizey,sizez,runtime,iter,(double)iter*(sizex-2)*(sizey-2)*(sizez-2)/runtime/1000000.);
  
  LIKWID_MARKER_CLOSE;
  return 0;
}

